# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import logging
import pytest
import inspect
import threading
import concurrent.futures
from tests.common.pipeline.helpers import (
    all_except,
    make_mock_op_or_event,
    assert_callback_failed,
    get_arg_count,
    add_mock_method_waiter,
    StageTestBase,
    StageRunOpTestBase,
    StageHandlePipelineEventTestBase,
)
from azure.iot.device.common.pipeline.pipeline_stages_base import PipelineStage, PipelineRootStage
from tests.common.pipeline.pipeline_data_object_test import add_instantiation_test
from azure.iot.device.common.pipeline import pipeline_thread, pipeline_exceptions
from azure.iot.device.common import handle_exceptions
from .fixtures import FakeOperation
from .helpers import assert_callback_succeeded

logging.basicConfig(level=logging.DEBUG)


def add_base_pipeline_stage_tests(
    cls,
    module,
    all_ops,
    handled_ops,
    all_events,
    handled_events,
    methods_that_enter_pipeline_thread=[],
    methods_that_can_run_in_any_thread=[],
    extra_initializer_defaults={},
    positional_arguments=[],
    keyword_arguments={},
):
    """
    Add all of the "basic" tests for validating a pipeline stage.  This includes tests for
    instantiation and tests for properly handling "unhandled" operations and events".
    """

    add_instantiation_test(
        cls=cls,
        module=module,
        defaults={"name": cls.__name__, "next": None, "previous": None, "pipeline_root": None},
        extra_defaults=extra_initializer_defaults,
        positional_arguments=positional_arguments,
        keyword_arguments=keyword_arguments,
    )
    _add_unknown_ops_tests(cls=cls, module=module, all_ops=all_ops, handled_ops=handled_ops)
    _add_unknown_events_tests(
        cls=cls, module=module, all_events=all_events, handled_events=handled_events
    )
    _add_pipeline_thread_tests(
        cls=cls,
        module=module,
        methods_that_enter_pipeline_thread=methods_that_enter_pipeline_thread,
        methods_that_can_run_in_any_thread=methods_that_can_run_in_any_thread,
    )
    _add_pipeline_flow_tests(cls=cls, module=module)


def _add_unknown_ops_tests(cls, module, all_ops, handled_ops):
    """
    Add tests for properly handling of "unknown operations," which are operations that aren't
    handled by a particular stage.  These operations should be passed down by any stage into
    the stages that follow.
    """
    unknown_ops = all_except(all_items=all_ops, items_to_exclude=handled_ops)

    @pytest.mark.describe("{} - .run_op() -- unknown and unhandled operations".format(cls.__name__))
    class LocalTestObject(StageRunOpTestBase):
        @pytest.fixture(params=unknown_ops)
        def op(self, request, mocker):
            op = make_mock_op_or_event(request.param)
            op.callback = mocker.MagicMock()
            add_mock_method_waiter(op, "callback")
            return op

        @pytest.fixture
        def stage(self):
            if cls == PipelineRootStage:
                return cls(None)
            else:
                return cls()

        @pytest.mark.it("Passes unknown operation down to the next stage")
        def test_passes_op_to_next_stage(self, mocker, op, stage):
            mocker.spy(stage, "send_op_down")
            stage.run_op(op)
            assert stage.send_op_down.call_count == 1
            assert stage.send_op_down.call_args == mocker.call(op)

    setattr(module, "Test{}UnknownOps".format(cls.__name__), LocalTestObject)


def _add_unknown_events_tests(cls, module, all_events, handled_events):
    """
    Add tests for properly handling of "unknown events," which are events that aren't
    handled by a particular stage.  These operations should be passed up by any stage into
    the stages that proceed it..
    """

    unknown_events = all_except(all_items=all_events, items_to_exclude=handled_events)

    if not unknown_events:
        return

    @pytest.mark.describe(
        "{} - .handle_pipeline_event() -- unknown and unhandled events".format(cls.__name__)
    )
    class LocalTestObject(StageHandlePipelineEventTestBase):
        @pytest.fixture(params=unknown_events)
        def event(self, request):
            return make_mock_op_or_event(request.param)

        @pytest.fixture
        def stage(self):
            return cls()

        @pytest.mark.it("Passes unknown event to previous stage")
        def test_passes_event_to_previous_stage(self, stage, event, mocker):
            mocker.spy(stage, "send_event_up")
            stage.handle_pipeline_event(event)

            assert stage.send_event_up.call_count == 1
            assert stage.send_event_up.call_args == mocker.call(event)

    setattr(module, "Test{}UnknownEvents".format(cls.__name__), LocalTestObject)


class ThreadLaunchedError(Exception):
    pass


def _add_pipeline_thread_tests(
    cls, module, methods_that_enter_pipeline_thread, methods_that_can_run_in_any_thread
):
    def does_method_assert_pipeline_thread(method_name):
        if method_name.startswith("__"):
            return False
        elif method_name in methods_that_enter_pipeline_thread:
            return False
        elif method_name in methods_that_can_run_in_any_thread:
            return False
        else:
            return True

    methods_that_assert_pipeline_thread = [
        x[0]
        for x in inspect.getmembers(cls, inspect.isfunction)
        if does_method_assert_pipeline_thread(x[0])
    ]

    @pytest.mark.describe("{} - Pipeline threading".format(cls.__name__))
    class LocalTestObject(object):
        @pytest.fixture
        def stage(self):
            # BKTODO: Make this more generic
            if cls == PipelineRootStage:
                return cls(None)
            else:
                return cls()

        @pytest.mark.parametrize("method_name", methods_that_assert_pipeline_thread)
        @pytest.mark.it("Enforces use of the pipeline thread when calling method")
        def test_asserts_in_pipeline(self, stage, method_name, fake_non_pipeline_thread):
            func = getattr(stage, method_name)
            args = [None for i in (range(get_arg_count(func) - 1))]
            with pytest.raises(AssertionError):
                func(*args)

        if methods_that_enter_pipeline_thread:

            @pytest.mark.parametrize("method_name", methods_that_enter_pipeline_thread)
            @pytest.mark.it("Automatically enters the pipeline thread when calling method")
            def test_enters_pipeline(self, mocker, stage, method_name, fake_non_pipeline_thread):
                func = getattr(stage, method_name)
                args = [None for i in (range(get_arg_count(func) - 1))]

                #
                # We take a bit of a roundabout way to verify that the functuion enters the
                # pipeline executor:
                #
                # 1. we verify that the method got the pipeline executor
                # 2. we verify that the method invoked _something_ on the pipeline executor
                #
                # It's not perfect, but it's good enough.
                #
                # We do this because:
                # 1. We don't have the exact right args to run the method and we don't want
                #    to add the complexity to get the right args in this test.
                # 2. We can't replace the wrapped method with a mock, AFAIK.
                #
                pipeline_executor = pipeline_thread._get_named_executor("pipeline")
                mocker.patch.object(pipeline_executor, "submit")
                pipeline_executor.submit.side_effect = ThreadLaunchedError
                mocker.spy(pipeline_thread, "_get_named_executor")

                # If the method calls submit on some executor, it will raise a ThreadLaunchedError
                with pytest.raises(ThreadLaunchedError):
                    func(*args)

                # now verify that the code got the pipeline executor and verify that it used that
                # executor to launch something.
                assert pipeline_thread._get_named_executor.call_count == 1
                assert pipeline_thread._get_named_executor.call_args[0][0] == "pipeline"
                assert pipeline_executor.submit.call_count == 1

    setattr(module, "Test{}PipelineThreading".format(cls.__name__), LocalTestObject)


def _add_pipeline_flow_tests(cls, module):
    class PipelineFlowTestBase(object):
        @pytest.fixture
        def stage(self, mocker):
            if cls == PipelineRootStage:
                stage = cls(None)
            else:
                stage = cls()
            stage.next = mocker.MagicMock()
            stage.previous = mocker.MagicMock()
            return stage

    @pytest.mark.describe("{} - .send_worker_op_down()".format(cls.__name__))
    class SendWorkerOpDownTests(PipelineFlowTestBase):
        @pytest.fixture
        def arbitrary_worker_op(self, mocker):
            op = FakeOperation(callback=mocker.MagicMock())
            op.name = "arbitrary_worker_op"
            return op

        @pytest.mark.it("Sends the worker operation to down to the next stage")
        def test_sends_worker_op_down(self, mocker, stage, arbitrary_op, arbitrary_worker_op):
            mocker.spy(stage, "send_op_down")

            stage.send_worker_op_down(worker_op=arbitrary_worker_op, op=arbitrary_op)

            assert stage.send_op_down.call_count == 1
            assert stage.send_op_down.call_args == mocker.call(arbitrary_worker_op)

        @pytest.mark.it(
            "Completes the original operation successfully upon successful completion of the worker operation"
        )
        def test_completes_original_operation_w_success(
            self, mocker, stage, arbitrary_op, arbitrary_worker_op
        ):
            mocker.spy(stage, "complete_op")

            stage.send_worker_op_down(worker_op=arbitrary_worker_op, op=arbitrary_op)
            assert stage.complete_op.call_count == 0
            arbitrary_worker_op.callback(arbitrary_worker_op, None)  # successful completion

            assert stage.complete_op.call_count == 1
            assert stage.complete_op.call_args == mocker.call(arbitrary_op, error=None)

        @pytest.mark.it(
            "Completes the original operation with the error from the worker op, if the worker op is completed with error"
        )
        def test_completes_original_operation_w_failure(
            self, mocker, stage, arbitrary_op, arbitrary_worker_op, arbitrary_exception
        ):
            mocker.spy(stage, "complete_op")

            stage.send_worker_op_down(worker_op=arbitrary_worker_op, op=arbitrary_op)
            assert stage.complete_op.call_count == 0
            arbitrary_worker_op.callback(
                arbitrary_worker_op, arbitrary_exception
            )  # successful completion

            assert stage.complete_op.call_count == 1
            assert stage.complete_op.call_args == mocker.call(
                arbitrary_op, error=arbitrary_exception
            )

    @pytest.mark.describe("{} - .send_op_down()".format(cls.__name__))
    class SendOpDownTests(PipelineFlowTestBase):
        @pytest.mark.it("Completes the op with failure if there is no next stage")
        def test_fails_op_when_no_next_stage(self, mocker, stage, arbitrary_op):
            mocker.spy(stage, "complete_op")
            stage.next = None

            stage.send_op_down(arbitrary_op)

            assert stage.complete_op.call_count == 1
            assert stage.complete_op.call_args[0][0] is arbitrary_op
            assert (
                type(stage.complete_op.call_args[1]["error"]) == pipeline_exceptions.PipelineError
            )

        @pytest.mark.it("Passes the op to the next stage's .run_op() method")
        def test_passes_op_to_next_stage(self, mocker, stage, arbitrary_op):
            stage.send_op_down(arbitrary_op)
            assert stage.next.run_op.call_count == 1
            assert stage.next.run_op.call_args == mocker.call(arbitrary_op)

    @pytest.mark.describe("{} - .complete_op()".format(cls.__name__))
    class CompleteOpTests(PipelineFlowTestBase):
        @pytest.mark.it(
            "Marks the operation as completed and sends it up the pipeline if the operation is not yet completed"
        )
        @pytest.mark.parametrize(
            "error",
            [
                pytest.param(None, id="Complete successfully"),
                pytest.param(Exception(), id="Complete with failure due to error"),
            ],
        )
        def test_uncompleted_op(self, mocker, stage, arbitrary_op, error):
            mocker.spy(stage, "send_completed_op_up")

            stage.complete_op(arbitrary_op, error)

            assert arbitrary_op.completed is True
            assert stage.send_completed_op_up.call_count == 1
            assert stage.send_completed_op_up.call_args == mocker.call(arbitrary_op, error)

        @pytest.mark.it(
            "Sends a PipelineError to the background exception handler instead of sending the operation up the pipeline, if the operation is already completed"
        )
        @pytest.mark.parametrize(
            "error",
            [
                pytest.param(None, id="Complete successfully"),
                pytest.param(Exception(), id="Complete with failure due to error"),
            ],
        )
        def test_op_already_completed(self, mocker, stage, arbitrary_op, error):
            mocker.spy(handle_exceptions, "handle_background_exception")
            mocker.spy(stage, "send_completed_op_up")
            arbitrary_op.completed = True

            stage.complete_op(arbitrary_op, error)

            assert handle_exceptions.handle_background_exception.call_count == 1
            assert (
                type(handle_exceptions.handle_background_exception.call_args[0][0])
                == pipeline_exceptions.PipelineError
            )
            assert stage.send_completed_op_up.call_count == 0

        @pytest.mark.it(
            "Defaults the 'error' parameter to None (i.e. Successful completion) if not specified"
        )
        def test_no_error_specified(self, mocker, stage, arbitrary_op):
            mocker.spy(stage, "send_completed_op_up")

            stage.complete_op(arbitrary_op)

            assert arbitrary_op.completed is True
            assert stage.send_completed_op_up.call_count == 1
            assert stage.send_completed_op_up.call_args == mocker.call(arbitrary_op, None)

    @pytest.mark.describe("{} - .send_completed_op_up()".format(cls.__name__))
    class SendCompletedOpUpTests(PipelineFlowTestBase):
        @pytest.mark.it("Calls the op callback on success")
        def test_calls_callback_on_success(self, stage, arbitrary_op):
            arbitrary_op.completed = True
            stage.send_completed_op_up(arbitrary_op)
            assert_callback_succeeded(arbitrary_op)

        @pytest.mark.it("Calls the op callback on failure")
        def test_calls_callback_on_error(self, stage, arbitrary_op, arbitrary_exception):
            arbitrary_op.completed = True
            stage.send_completed_op_up(arbitrary_op, error=arbitrary_exception)
            assert_callback_failed(op=arbitrary_op, error=arbitrary_exception)

        @pytest.mark.it(
            "Calls the callback with a PipelineError if the operation has not been completed"
        )
        def test_error_if_not_completed(self, stage, arbitrary_op):
            with pytest.raises(pipeline_exceptions.PipelineError):
                stage.send_completed_op_up(arbitrary_op)

        @pytest.mark.it(
            "Handles Exceptions raised in the operation callback by passing them to the background exception handler"
        )
        def test_op_callback_raises_exception(
            self, stage, arbitrary_op, arbitrary_exception, mocker, unhandled_error_handler
        ):
            arbitrary_op.callback = mocker.Mock(side_effect=arbitrary_exception)
            arbitrary_op.completed = True
            stage.send_completed_op_up(arbitrary_op)
            assert arbitrary_op.callback.call_count == 1
            assert arbitrary_op.callback.call_args == mocker.call(arbitrary_op, error=None)
            assert unhandled_error_handler.call_count == 1
            assert unhandled_error_handler.call_args == mocker.call(arbitrary_exception)

        @pytest.mark.it("Allows any BaseExceptions raised in the operation callback to propagate")
        def test_op_callback_raises_base_exception(
            self, stage, arbitrary_op, arbitrary_base_exception, mocker
        ):
            arbitrary_op.callback = mocker.Mock(side_effect=arbitrary_base_exception)
            arbitrary_op.completed = True
            with pytest.raises(arbitrary_base_exception.__class__):
                stage.send_completed_op_up(arbitrary_op)

    @pytest.mark.describe("{} - .send_op_down_and_intercept_return()".format(cls.__name__))
    class SendOpDownAndInterceptReturnTests(PipelineFlowTestBase):
        @pytest.mark.it(
            "Sends the operation down the pipeline after replacing its original callback"
        )
        def test_sends_op_down(self, stage, arbitrary_op, mocker):
            intercepted_return = mocker.MagicMock()
            original_callback = arbitrary_op.callback
            mocker.spy(stage, "send_op_down")

            stage.send_op_down_and_intercept_return(arbitrary_op, intercepted_return)

            assert stage.send_op_down.call_count == 1
            assert stage.send_op_down.call_args == mocker.call(arbitrary_op)
            assert arbitrary_op.callback is not original_callback

        @pytest.mark.it("Calls the intercepted_return function upon completion of the operation")
        @pytest.mark.parametrize(
            "error",
            [
                pytest.param(None, id="Completed successfully"),
                pytest.param(Exception(), id="Completed with failure due to error"),
            ],
        )
        def test_intercept_called(self, stage, arbitrary_op, error, mocker):
            intercepted_return = mocker.MagicMock()
            original_callback = arbitrary_op.callback

            stage.send_op_down_and_intercept_return(arbitrary_op, intercepted_return)
            arbitrary_op.callback(arbitrary_op, error)  # Complete op

            assert intercepted_return.call_count == 1
            assert intercepted_return.call_args == mocker.call(op=arbitrary_op, error=error)
            assert original_callback.call_count == 0  # Original cb not called

        @pytest.mark.it(
            "Resets the original callback to the operation upon completion of the operation"
        )
        @pytest.mark.parametrize(
            "error",
            [
                pytest.param(None, id="Completed successfully"),
                pytest.param(Exception(), id="Completed with failure due to error"),
            ],
        )
        def test_callback_reset(self, stage, arbitrary_op, error, mocker):
            intercepted_return = mocker.MagicMock()
            original_callback = arbitrary_op.callback

            stage.send_op_down_and_intercept_return(arbitrary_op, intercepted_return)
            assert arbitrary_op.callback is not original_callback
            arbitrary_op.callback(arbitrary_op, error)  # Complete op

            assert arbitrary_op.callback is original_callback

    @pytest.mark.describe("{} - .send_event_up()".format(cls.__name__))
    class SendEventUpTests(PipelineFlowTestBase):
        @pytest.mark.it(
            "Passes the event up to the previous stage's .handle_pipeline_event() method"
        )
        def test_calls_handle_pipeline_event(self, stage, arbitrary_event, mocker):
            stage.send_event_up(arbitrary_event)
            assert stage.previous.handle_pipeline_event.call_count == 1
            assert stage.previous.handle_pipeline_event.call_args == mocker.call(arbitrary_event)

        @pytest.mark.it(
            "Sends a PipelineError to the background exception handler instead of sending the event up the pipeline, if there is no previous pipeline stage"
        )
        def test_no_previous_stage(self, stage, arbitrary_event, mocker):
            stage.previous = None
            mocker.spy(handle_exceptions, "handle_background_exception")

            stage.send_event_up(arbitrary_event)

            assert handle_exceptions.handle_background_exception.call_count == 1
            assert (
                type(handle_exceptions.handle_background_exception.call_args[0][0])
                == pipeline_exceptions.PipelineError
            )

    setattr(module, "Test{}SendWorkerOpDown".format(cls.__name__), SendWorkerOpDownTests)
    setattr(module, "Test{}SendOpDown".format(cls.__name__), SendOpDownTests)
    setattr(module, "Test{}CompleteOp".format(cls.__name__), CompleteOpTests)
    setattr(module, "Test{}SendCompletedOpUp".format(cls.__name__), SendCompletedOpUpTests)
    setattr(
        module,
        "Test{}SendOpDwonAndInterceptReturn".format(cls.__name__),
        SendOpDownAndInterceptReturnTests,
    )
    setattr(module, "Test{}SendEventUp".format(cls.__name__), SendEventUpTests)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import inspect
import pytest
import functools
from threading import Event
from azure.iot.device.common import handle_exceptions
from azure.iot.device.common.pipeline import (
    pipeline_events_base,
    pipeline_ops_base,
    pipeline_stages_base,
    pipeline_events_mqtt,
    pipeline_ops_mqtt,
    config,
)

try:
    from inspect import getfullargspec as getargspec
except ImportError:
    from inspect import getargspec

all_common_ops = [
    pipeline_ops_base.ConnectOperation,
    pipeline_ops_base.ReconnectOperation,
    pipeline_ops_base.DisconnectOperation,
    pipeline_ops_base.EnableFeatureOperation,
    pipeline_ops_base.DisableFeatureOperation,
    pipeline_ops_base.UpdateSasTokenOperation,
    pipeline_ops_base.RequestAndResponseOperation,
    pipeline_ops_base.RequestOperation,
    pipeline_ops_mqtt.SetMQTTConnectionArgsOperation,
    pipeline_ops_mqtt.MQTTPublishOperation,
    pipeline_ops_mqtt.MQTTSubscribeOperation,
    pipeline_ops_mqtt.MQTTUnsubscribeOperation,
]

all_common_events = [pipeline_events_mqtt.IncomingMQTTMessageEvent]


def all_except(all_items, items_to_exclude):
    """
    helper function to return a new list with all ops that are in the first list
    and not in the second list.

    :param list all_items: list of all operations or events
    :param list items_to_exclude: ops or events to exclude
    """
    return [x for x in all_items if x not in items_to_exclude]


# TODO: Remove this definition
class StageTestBase(object):
    @pytest.fixture(autouse=True)
    def stage_base_configuration(self, stage, mocker):
        """
        This fixture configures the stage for testing.  This is automatically
        applied, so it will be called before your test runs, but it's not
        guaranteed to be called before any other fixtures run.  If you have
        a fixture that needs to rely on the stage being configured, then
        you have to add a manual dependency inside that fixture (like we do in
        next_stage_succeeds_all_ops below)
        """

        class NextStageForTest(pipeline_stages_base.PipelineStage):
            def _execute_op(self, op):
                pass

        next = NextStageForTest()
        root = (
            pipeline_stages_base.PipelineRootStage(config.BasePipelineConfig())
            .append_stage(stage)
            .append_stage(next)
        )

        mocker.spy(stage, "_execute_op")
        mocker.spy(stage, "run_op")

        mocker.spy(next, "_execute_op")
        mocker.spy(next, "run_op")

        return root

    @pytest.fixture
    def next_stage_succeeds(self, stage, stage_base_configuration, mocker):
        stage.next._execute_op = stage.complete_op
        mocker.spy(stage.next, "_execute_op")

    @pytest.fixture
    def next_stage_raises_arbitrary_exception(
        self, stage, stage_base_configuration, mocker, arbitrary_exception
    ):
        stage.next._execute_op = mocker.MagicMock(side_effect=arbitrary_exception)

    @pytest.fixture
    def next_stage_raises_arbitrary_base_exception(
        self, stage, stage_base_configuration, mocker, arbitrary_base_exception
    ):
        stage.next._execute_op = mocker.MagicMock(side_effect=arbitrary_base_exception)


# TODO: refactor all .run_op() tests to use this
class StageRunOpTestBase(object):
    """All PipelineStage .run_op() tests should inherit from this base class.
    It provides basic tests for dealing with exceptions.
    """

    @pytest.mark.it(
        "Completes the operation with failure if an unexpected Exception is raised while executing the operation"
    )
    def test_completes_operation_with_error(self, mocker, stage, op, arbitrary_exception):
        stage._execute_op = mocker.MagicMock(side_effect=arbitrary_exception)
        mocker.spy(stage, "complete_op")

        stage.run_op(op)

        assert stage.complete_op.call_count == 1
        assert stage.complete_op.call_args == mocker.call(op, error=arbitrary_exception)

    @pytest.mark.it(
        "Allows any BaseException that was raised during execution of the operation to propogate"
    )
    def test_base_exception_propogates(self, mocker, stage, op, arbitrary_base_exception):
        stage._execute_op = mocker.MagicMock(side_effect=arbitrary_base_exception)

        with pytest.raises(arbitrary_base_exception.__class__) as e_info:
            stage.run_op(op)
        assert e_info.value is arbitrary_base_exception


# TODO: refactor all .handle_pipeline_event() tests to use this
class StageHandlePipelineEventTestBase(object):
    """All PipelineStage .handle_pipeline_event() tests should inherit from this base class.
    It provides basic tests for dealing with exceptions.
    """

    @pytest.mark.it(
        "Sends any unexpected Exceptions raised during handling of the event to the background exception handler"
    )
    def test_uses_background_exception_handler(self, mocker, stage, event, arbitrary_exception):
        stage._handle_pipeline_event = mocker.MagicMock(side_effect=arbitrary_exception)
        mocker.spy(handle_exceptions, "handle_background_exception")

        stage.handle_pipeline_event(event)

        assert handle_exceptions.handle_background_exception.call_count == 1
        assert handle_exceptions.handle_background_exception.call_args == mocker.call(
            arbitrary_exception
        )

    @pytest.mark.it("Allows any BaseException raised during handling of the event to propogate")
    def test_base_exception_propogates(self, mocker, stage, event, arbitrary_base_exception):
        stage._handle_pipeline_event = mocker.MagicMock(side_effect=arbitrary_base_exception)

        with pytest.raises(arbitrary_base_exception.__class__) as e_info:
            stage.handle_pipeline_event(event)
        assert e_info.value is arbitrary_base_exception


def assert_callback_succeeded(op, callback=None):
    if not callback:
        callback = op.callback
    try:
        # if the callback has a __wrapped__ attribute, that means that the
        # pipeline added a wrapper around the callback, so we want to look
        # at the original function instead of the wrapped function.
        callback = callback.__wrapped__
    except AttributeError:
        pass
    assert callback.call_count == 1
    callback_op_arg = callback.call_args[0][0]
    assert callback_op_arg == op
    callback_error_arg = callback.call_args[1]["error"]
    assert callback_error_arg is None


def assert_callback_failed(op, callback=None, error=None):
    if not callback:
        callback = op.callback
    try:
        # if the callback has a __wrapped__ attribute, that means that the
        # pipeline added a wrapper around the callback, so we want to look
        # at the original function instead of the wrapped function.
        callback = callback.__wrapped__
    except AttributeError:
        pass
    assert callback.call_count == 1
    callback_op_arg = callback.call_args[0][0]
    assert callback_op_arg == op

    callback_error_arg = callback.call_args[1]["error"]
    if error:
        if isinstance(error, type):
            assert callback_error_arg.__class__ == error
        else:
            assert callback_error_arg is error
    else:
        assert callback_error_arg is not None


def get_arg_count(fn):
    """
    return the number of arguments (args) passed into a
    particular function.  Returned value does not include kwargs.
    """
    try:
        # if __wrapped__ is set, we're looking at a decorated function
        # Functools.wraps doesn't copy arg metadata, so we need to
        # get argument count from the wrapped function instead.
        return len(getargspec(fn.__wrapped__).args)
    except AttributeError:
        return len(getargspec(fn).args)


def make_mock_op_or_event(cls):
    args = [None for i in (range(get_arg_count(cls.__init__) - 1))]
    return cls(*args)


def add_mock_method_waiter(obj, method_name):
    """
    For mock methods, add "wait_for_xxx_to_be_called" and "wait_for_xxx_to_not_be_called"
    helper functions on the object.  This is very handy for methods that get called by
    another thread, when you want your test functions to wait until the other thread is
    able to call the method without using a sleep call.
    """
    method_called = Event()

    def signal_method_called(*args, **kwargs):
        method_called.set()

    def wait_for_method_to_be_called():
        method_called.wait(0.1)
        assert method_called.isSet()
        method_called.clear()

    def wait_for_method_to_not_be_called():
        method_called.wait(0.1)
        assert not method_called.isSet()

    getattr(obj, method_name).side_effect = signal_method_called
    setattr(obj, "wait_for_{}_to_be_called".format(method_name), wait_for_method_to_be_called)
    setattr(
        obj, "wait_for_{}_to_not_be_called".format(method_name), wait_for_method_to_not_be_called
    )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import logging
import pytest
import sys
import six
import threading
from six.moves import queue
from azure.iot.device.common import transport_exceptions
from azure.iot.device.common.pipeline import (
    pipeline_stages_base,
    pipeline_ops_base,
    pipeline_ops_mqtt,
    pipeline_events_base,
    pipeline_exceptions,
)
from tests.common.pipeline.helpers import (
    assert_callback_failed,
    assert_callback_succeeded,
    all_common_ops,
    all_common_events,
    StageTestBase,
    StageRunOpTestBase,
    StageHandlePipelineEventTestBase,
    all_except,
    make_mock_op_or_event,
)
from tests.common.pipeline import pipeline_stage_test

this_module = sys.modules[__name__]
logging.basicConfig(level=logging.DEBUG)


# This fixture makes it look like all test in this file  tests are running
# inside the pipeline thread.  Because this is an autouse fixture, we
# manually add it to the individual test.py files that need it.  If,
# instead, we had added it to some conftest.py, it would be applied to
# every tests in every file and we don't want that.
@pytest.fixture(autouse=True)
def apply_fake_pipeline_thread(fake_pipeline_thread):
    pass


# Workaround for flake8.  A class with this name is actually created inside
# add_base_pipeline_stage_test, but flake8 doesn't know that
class TestPipelineRootStagePipelineThreading:
    pass


pipeline_stage_test.add_base_pipeline_stage_tests(
    cls=pipeline_stages_base.PipelineRootStage,
    module=this_module,
    all_ops=all_common_ops,
    handled_ops=[],
    all_events=all_common_events,
    handled_events=all_common_events,
    methods_that_can_run_in_any_thread=["append_stage", "run_op"],
    extra_initializer_defaults={
        "on_pipeline_event_handler": None,
        "on_connected_handler": None,
        "on_disconnected_handler": None,
        "connected": False,
    },
    positional_arguments=["pipeline_configuration"],
)


@pytest.mark.it("Calls operation callback in callback thread")
def _test_pipeline_root_runs_callback_in_callback_thread(self, stage, mocker):
    # the stage fixture comes from the TestPipelineRootStagePipelineThreading object that
    # this test method gets added to, so it's a PipelineRootStage object
    stage.pipeline_root = stage
    callback_called = threading.Event()

    def callback(op, error):
        assert threading.current_thread().name == "callback"
        callback_called.set()

    op = pipeline_ops_base.ConnectOperation(callback=callback)
    stage.run_op(op)
    callback_called.wait()


@pytest.mark.it("Runs operation in pipeline thread")
def _test_pipeline_root_runs_operation_in_pipeline_thread(
    self, mocker, stage, arbitrary_op, fake_non_pipeline_thread
):
    # the stage fixture comes from the TestPipelineRootStagePipelineThreading object that
    # this test method gets added to, so it's a PipelineRootStage object
    assert threading.current_thread().name != "pipeline"

    def mock_execute_op(self, op):
        print("mock_execute_op called")
        assert threading.current_thread().name == "pipeline"
        op.callback(op)

    mock_execute_op = mocker.MagicMock(mock_execute_op)
    stage._execute_op = mock_execute_op

    stage.run_op(arbitrary_op)
    assert mock_execute_op.call_count == 1


@pytest.mark.it("Calls on_connected_handler in callback thread")
def _test_pipeline_root_runs_on_connected_in_callback_thread(self, stage, mocker):
    stage.pipeline_root = stage
    callback_called = threading.Event()

    def callback(*arg, **argv):
        assert threading.current_thread().name == "callback"
        callback_called.set()

    stage.on_connected_handler = callback

    stage.on_connected()
    callback_called.wait()


@pytest.mark.it("Calls on_disconnected_handler in callback thread")
def _test_pipeline_root_runs_on_disconnected_in_callback_thread(self, stage, mocker):
    stage.pipeline_root = stage
    callback_called = threading.Event()

    def callback(*arg, **argv):
        assert threading.current_thread().name == "callback"
        callback_called.set()

    stage.on_disconnected_handler = callback

    stage.on_disconnected()
    callback_called.wait()


@pytest.mark.it("Calls on_event_received_handler in callback thread")
def _test_pipeline_root_runs_on_event_received_in_callback_thread(
    self, stage, mocker, arbitrary_event
):
    stage.pipeline_root = stage
    callback_called = threading.Event()

    def callback(*arg, **argv):
        assert threading.current_thread().name == "callback"
        callback_called.set()

    stage.on_pipeline_event_handler = callback

    stage.handle_pipeline_event(arbitrary_event)
    callback_called.wait()


TestPipelineRootStagePipelineThreading.test_runs_callback_in_callback_thread = (
    _test_pipeline_root_runs_callback_in_callback_thread
)
TestPipelineRootStagePipelineThreading.test_runs_operation_in_pipeline_thread = (
    _test_pipeline_root_runs_operation_in_pipeline_thread
)
TestPipelineRootStagePipelineThreading.test_pipeline_root_runs_on_connected_in_callback_thread = (
    _test_pipeline_root_runs_on_connected_in_callback_thread
)
TestPipelineRootStagePipelineThreading.test_pipeline_root_runs_on_disconnected_in_callback_thread = (
    _test_pipeline_root_runs_on_disconnected_in_callback_thread
)
TestPipelineRootStagePipelineThreading.test_pipeline_root_runs_on_event_received_in_callback_thread = (
    _test_pipeline_root_runs_on_event_received_in_callback_thread
)

pipeline_stage_test.add_base_pipeline_stage_tests(
    cls=pipeline_stages_base.AutoConnectStage,
    module=this_module,
    all_ops=all_common_ops,
    handled_ops=[
        pipeline_ops_mqtt.MQTTPublishOperation,
        pipeline_ops_mqtt.MQTTSubscribeOperation,
        pipeline_ops_mqtt.MQTTUnsubscribeOperation,
    ],
    all_events=all_common_events,
    handled_events=[],
)

fake_topic = "__fake_topic__"
fake_payload = "__fake_payload__"
ops_that_cause_connection = [
    {
        "op_class": pipeline_ops_mqtt.MQTTPublishOperation,
        "op_init_kwargs": {"topic": fake_topic, "payload": fake_payload, "callback": 1},
    },
    {
        "op_class": pipeline_ops_mqtt.MQTTSubscribeOperation,
        "op_init_kwargs": {"topic": fake_topic, "callback": 1},
    },
    {
        "op_class": pipeline_ops_mqtt.MQTTUnsubscribeOperation,
        "op_init_kwargs": {"topic": fake_topic, "callback": 1},
    },
]


@pytest.mark.parametrize(
    "params",
    ops_that_cause_connection,
    ids=[x["op_class"].__name__ for x in ops_that_cause_connection],
)
@pytest.mark.describe(
    "AutoConnectStage - .run_op() -- called with operation that causes a connection to be established"
)
class TestAutoConnectStageRunOp(StageTestBase):
    @pytest.fixture
    def op(self, mocker, params):
        op = params["op_class"](**params["op_init_kwargs"])
        op.callback = mocker.MagicMock()
        return op

    @pytest.fixture
    def stage(self):
        return pipeline_stages_base.AutoConnectStage()

    @pytest.mark.it("Passes the operation down the pipline when the transport is already connected")
    def test_operation_alrady_connected(self, params, op, stage):
        stage.pipeline_root.connected = True

        stage.run_op(op)

        assert stage.next.run_op.call_count == 1
        assert stage.next.run_op.call_args[0][0] == op

    @pytest.mark.it(
        "Sends a ConnectOperation instead of the op down the pipeline if the transport is not connected"
    )
    def test_sends_connect(self, params, op, stage):
        stage.pipeline_root.connected = False

        stage.run_op(op)

        assert stage.next.run_op.call_count == 1
        assert isinstance(stage.next.run_op.call_args[0][0], pipeline_ops_base.ConnectOperation)

    @pytest.mark.it(
        "Calls the op's callback with the error from the ConnectOperation if that operation fails"
    )
    def test_connect_failure(self, params, op, stage, arbitrary_exception):
        stage.pipeline_root.connected = False

        stage.run_op(op)
        connect_op = stage.next.run_op.call_args[0][0]
        stage.next.complete_op(connect_op, error=arbitrary_exception)

        assert_callback_failed(op=op, error=arbitrary_exception)

    @pytest.mark.it("Waits for the ConnectOperation to complete before pasing the operation down")
    def test_connect_success(self, params, op, stage):
        stage.pipeline_root.connected = False

        stage.run_op(op)
        assert stage.next.run_op.call_count == 1
        connect_op = stage.next.run_op.call_args[0][0]
        stage.next.complete_op(connect_op)

        assert stage.next.run_op.call_count == 2
        assert stage.next.run_op.call_args[0][0] == op

    @pytest.mark.it("calls the op's callback when the operation is complete after connecting")
    def test_operation_complete(self, params, op, stage):
        stage.pipeline_root.connected = False

        stage.run_op(op)
        connect_op = stage.next.run_op.call_args[0][0]
        stage.next.complete_op(connect_op)

        stage.next.complete_op(op)
        assert_callback_succeeded(op=op)

    @pytest.mark.it("calls the op's callback when the operation fails after connecting")
    def test_operation_fails(self, params, op, stage, arbitrary_exception):
        stage.pipeline_root.connected = False

        stage.run_op(op)
        connect_op = stage.next.run_op.call_args[0][0]
        stage.next.complete_op(connect_op)
        stage.next.complete_op(op, error=arbitrary_exception)

        assert_callback_failed(op=op, error=arbitrary_exception)


pipeline_stage_test.add_base_pipeline_stage_tests(
    cls=pipeline_stages_base.ConnectionLockStage,
    module=this_module,
    all_ops=all_common_ops,
    handled_ops=[
        pipeline_ops_base.ConnectOperation,
        pipeline_ops_base.DisconnectOperation,
        pipeline_ops_base.ReconnectOperation,
    ],
    all_events=all_common_events,
    handled_events=[],
    extra_initializer_defaults={"blocked": False, "queue": queue.Queue},
)

connection_ops = [
    {"op_class": pipeline_ops_base.ConnectOperation, "connected_flag_required_to_run": False},
    {"op_class": pipeline_ops_base.DisconnectOperation, "connected_flag_required_to_run": True},
    {"op_class": pipeline_ops_base.ReconnectOperation, "connected_flag_required_to_run": True},
]


class FakeOperation(pipeline_ops_base.PipelineOperation):
    pass


@pytest.mark.describe(
    "ConnectionLockStage - .run_op() -- called with an operation that connects, disconnects, or reconnects"
)
class TestSerializeConnectOpStageRunOp(StageTestBase):
    @pytest.fixture
    def stage(self):
        return pipeline_stages_base.ConnectionLockStage()

    @pytest.fixture
    def connection_op(self, mocker, params):
        return params["op_class"](callback=mocker.MagicMock())

    @pytest.fixture
    def fake_op(self, mocker):
        return FakeOperation(callback=mocker.MagicMock())

    @pytest.fixture
    def fake_ops(self, mocker):
        return [
            FakeOperation(callback=mocker.MagicMock()),
            FakeOperation(callback=mocker.MagicMock()),
            FakeOperation(callback=mocker.MagicMock()),
        ]

    @pytest.mark.it(
        "Immediately completes a ConnectOperation if the transport is already connected"
    )
    def test_connect_while_connected(self, stage, mocker):
        op = pipeline_ops_base.ConnectOperation(callback=mocker.MagicMock())
        stage.pipeline_root.connected = True
        stage.run_op(op)
        assert_callback_succeeded(op=op)

    @pytest.mark.it(
        "Immediately completes a DisconnectOperation if the transport is already disconnected"
    )
    def test_disconnect_while_disconnected(self, stage, mocker):
        op = pipeline_ops_base.DisconnectOperation(callback=mocker.MagicMock())
        stage.pipeline_root.connected = False
        stage.run_op(op)
        assert_callback_succeeded(op=op)

    @pytest.mark.it(
        "Immediately passes the operation down if an operation is not alrady being blocking the stage"
    )
    def test_passes_op_when_not_blocked(self, stage, mocker, fake_op):
        stage.run_op(fake_op)
        assert stage.next.run_op.call_count == 1
        assert stage.next.run_op.call_args[0][0] == fake_op

    @pytest.mark.parametrize(
        "params", connection_ops, ids=[x["op_class"].__name__ for x in connection_ops]
    )
    @pytest.mark.it(
        "Does not immediately pass the operation down if a different operation is currently blcking the stage"
    )
    def test_does_not_pass_op_if_blocked(self, params, stage, connection_op, fake_op):
        stage.pipeline_root.connected = params["connected_flag_required_to_run"]
        stage.run_op(connection_op)
        stage.run_op(fake_op)

        assert stage.next.run_op.call_count == 1
        assert stage.next.run_op.call_args[0][0] == connection_op

    @pytest.mark.parametrize(
        "params", connection_ops, ids=[x["op_class"].__name__ for x in connection_ops]
    )
    @pytest.mark.it(
        "Waits for the operation that is currently blocking the stage to complete before passing the op down"
    )
    def test_waits_for_serialized_op_to_complete_before_passing_blocked_op(
        self, params, stage, connection_op, fake_op
    ):
        stage.pipeline_root.connected = params["connected_flag_required_to_run"]
        stage.run_op(connection_op)
        stage.run_op(fake_op)
        stage.next.complete_op(connection_op)

        assert stage.next.run_op.call_count == 2
        assert stage.next.run_op.call_args[0][0] == fake_op

    @pytest.mark.parametrize(
        "params", connection_ops, ids=[x["op_class"].__name__ for x in connection_ops]
    )
    @pytest.mark.it("Fails the operation if the operation that previously blocked the stage fails")
    def test_fails_blocked_op_if_serialized_op_fails(
        self, params, stage, connection_op, fake_op, arbitrary_exception
    ):
        stage.pipeline_root.connected = params["connected_flag_required_to_run"]
        stage.run_op(connection_op)
        stage.run_op(fake_op)
        stage.next.complete_op(connection_op, error=arbitrary_exception)
        assert_callback_failed(op=fake_op, error=arbitrary_exception)

    @pytest.mark.parametrize(
        "params", connection_ops, ids=[x["op_class"].__name__ for x in connection_ops]
    )
    @pytest.mark.it(
        "Can pend multiple operations while waiting for an operation that is currently blocking the stage"
    )
    def test_blocks_multiple_ops(self, params, stage, connection_op, fake_ops):
        stage.pipeline_root.connected = params["connected_flag_required_to_run"]
        stage.run_op(connection_op)
        for op in fake_ops:
            stage.run_op(op)
        assert stage.next.run_op.call_count == 1

    @pytest.mark.parametrize(
        "params", connection_ops, ids=[x["op_class"].__name__ for x in connection_ops]
    )
    @pytest.mark.it(
        "Passes down all pending operations after the operation that previously blocked the stage completes successfully"
    )
    def test_unblocks_multiple_ops(self, params, stage, connection_op, fake_ops):
        stage.pipeline_root.connected = params["connected_flag_required_to_run"]
        stage.run_op(connection_op)
        for op in fake_ops:
            stage.run_op(op)

        stage.next.complete_op(connection_op)

        assert stage.next.run_op.call_count == 1 + len(fake_ops)

        # zip our ops and our calls together and make sure they match
        run_ops = zip(fake_ops, stage.next.run_op.call_args_list[1:])
        for run_op in run_ops:
            op = run_op[0]
            call_args = run_op[1]
            assert op == call_args[0][0]

    @pytest.mark.parametrize(
        "params", connection_ops, ids=[x["op_class"].__name__ for x in connection_ops]
    )
    @pytest.mark.it(
        "Fails all pending operations after the operation that previously blocked the stage fails"
    )
    def test_fails_multiple_ops(self, params, stage, connection_op, fake_ops, arbitrary_exception):
        stage.pipeline_root.connected = params["connected_flag_required_to_run"]
        stage.run_op(connection_op)
        for op in fake_ops:
            stage.run_op(op)

        stage.next.complete_op(connection_op, error=arbitrary_exception)

        for op in fake_ops:
            assert_callback_failed(op=op, error=arbitrary_exception)

    @pytest.mark.it(
        "Does not immediately pass down operations in the queue if an operation in the queue causes the stage to re-block"
    )
    def test_re_blocks_ops_from_queue(self, stage, mocker):
        first_connect = pipeline_ops_base.ConnectOperation(callback=mocker.MagicMock())
        first_fake_op = FakeOperation(callback=mocker.MagicMock())
        second_connect = pipeline_ops_base.ReconnectOperation(callback=mocker.MagicMock())
        second_fake_op = FakeOperation(callback=mocker.MagicMock())

        stage.run_op(first_connect)
        stage.run_op(first_fake_op)
        stage.run_op(second_connect)
        stage.run_op(second_fake_op)

        # at this point, ops are pended waiting for the first connect to complete.  Verify this and complete the connect.
        assert stage.next.run_op.call_count == 1
        assert stage.next.run_op.call_args[0][0] == first_connect
        stage.next.complete_op(first_connect)

        # The connect is complete.  This passes down first_fake_op and second_connect and second_fake_op gets pended waiting i
        # for second_connect to complete.
        # Note: this isn't ideal.  In a perfect world, second_connect wouldn't start until first_fake_op is complete, but we
        # dont have this logic in place yet.
        assert stage.next.run_op.call_count == 3
        assert stage.next.run_op.call_args_list[1][0][0] == first_fake_op
        assert stage.next.run_op.call_args_list[2][0][0] == second_connect

        # now, complete second_connect to give second_fake_op a chance to get passed down
        stage.next.complete_op(second_connect)
        assert stage.next.run_op.call_count == 4
        assert stage.next.run_op.call_args_list[3][0][0] == second_fake_op

    @pytest.mark.parametrize(
        "params",
        [
            pytest.param(
                {
                    "pre_connected_flag": True,
                    "first_connection_op": pipeline_ops_base.DisconnectOperation,
                    "mid_connect_flag": False,
                    "second_connection_op": pipeline_ops_base.DisconnectOperation,
                },
                id="Disconnect followed by Disconnect",
            ),
            pytest.param(
                {
                    "pre_connected_flag": False,
                    "first_connection_op": pipeline_ops_base.ConnectOperation,
                    "mid_connect_flag": True,
                    "second_connection_op": pipeline_ops_base.ConnectOperation,
                },
                id="Connect followed by Connect",
            ),
            pytest.param(
                {
                    "pre_connected_flag": True,
                    "first_connection_op": pipeline_ops_base.ReconnectOperation,
                    "mid_connect_flag": True,
                    "second_connection_op": pipeline_ops_base.ConnectOperation,
                },
                id="Reconnect followed by Connect",
            ),
        ],
    )
    @pytest.mark.it(
        "Immediately completes a second op which was waiting for a first op that succeeded"
    )
    def test_immediately_completes_second_op(self, stage, params, mocker):
        first_connection_op = params["first_connection_op"](mocker.MagicMock())
        second_connection_op = params["second_connection_op"](mocker.MagicMock())
        stage.pipeline_root.connected = params["pre_connected_flag"]

        stage.run_op(first_connection_op)
        stage.run_op(second_connection_op)

        # first_connection_op has been passed down.  second_connection_op is waiting for first disconnect to complete.
        assert stage.next.run_op.call_count == 1
        assert stage.next.run_op.call_args[0][0] == first_connection_op

        # complete first_connection_op
        stage.pipeline_root.connected = params["mid_connect_flag"]
        stage.next.complete_op(first_connection_op)

        # second connect_op should be completed without having been passed down.
        assert stage.next.run_op.call_count == 1
        assert_callback_succeeded(op=second_connection_op)


pipeline_stage_test.add_base_pipeline_stage_tests(
    cls=pipeline_stages_base.CoordinateRequestAndResponseStage,
    module=this_module,
    all_ops=all_common_ops,
    handled_ops=[pipeline_ops_base.RequestAndResponseOperation],
    all_events=all_common_events,
    handled_events=[pipeline_events_base.ResponseEvent],
    extra_initializer_defaults={"pending_responses": dict},
)


fake_request_type = "__fake_request_type__"
fake_method = "__fake_method__"
fake_resource_location = "__fake_resource_location__"
fake_request_body = "__fake_request_body__"
fake_status_code = "__fake_status_code__"
fake_response_body = "__fake_response_body__"
fake_request_id = "__fake_request_id__"


def make_fake_request_and_response(mocker):
    return pipeline_ops_base.RequestAndResponseOperation(
        request_type=fake_request_type,
        method=fake_method,
        resource_location=fake_resource_location,
        request_body=fake_request_body,
        callback=mocker.MagicMock(),
    )


@pytest.mark.describe(
    "CoordinateRequestAndResponse - .run_op() -- called with RequestAndResponseOperation"
)
class TestCoordinateRequestAndResponseSendIotRequestRunOp(StageTestBase):
    @pytest.fixture
    def op(self, mocker):
        return make_fake_request_and_response(mocker)

    @pytest.fixture
    def stage(self):
        return pipeline_stages_base.CoordinateRequestAndResponseStage()

    @pytest.mark.it(
        "Sends an RequestOperation op to the next stage with the same parameters and a newly allocated request_id"
    )
    def test_sends_op_and_validates_new_op(self, stage, op):
        stage.run_op(op)
        assert stage.next.run_op.call_count == 1
        new_op = stage.next.run_op.call_args[0][0]
        assert isinstance(new_op, pipeline_ops_base.RequestOperation)
        assert new_op.request_type == op.request_type
        assert new_op.method == op.method
        assert new_op.resource_location == op.resource_location
        assert new_op.request_body == op.request_body
        assert new_op.request_id

    @pytest.mark.it("Does not complete the SendIotRequestAndwaitForResponse op")
    def test_sends_op_and_verifies_no_response(self, stage, op):
        stage.run_op(op)
        assert op.callback.call_count == 0

    @pytest.mark.it("Fails RequestAndResponseOperation if there is no next stage")
    def test_no_next_stage(self, stage, op):
        stage.next = None
        stage.run_op(op)
        assert_callback_failed(op=op)

    @pytest.mark.it("Generates a new request_id for every operation")
    def test_sends_two_ops_and_validates_request_id(self, stage, op, mocker):
        op2 = make_fake_request_and_response(mocker)
        stage.run_op(op)
        stage.run_op(op2)
        assert stage.next.run_op.call_count == 2
        new_op = stage.next.run_op.call_args_list[0][0][0]
        new_op2 = stage.next.run_op.call_args_list[1][0][0]
        assert new_op.request_id != new_op2.request_id

    @pytest.mark.it(
        "Fails RequestAndResponseOperation if an Exception is raised in the RequestOperation op"
    )
    def test_new_op_raises_exception(self, stage, op, mocker, arbitrary_exception):
        stage.next._execute_op = mocker.Mock(side_effect=arbitrary_exception)
        stage.run_op(op)
        assert_callback_failed(op=op)

    @pytest.mark.it("Allows BaseExceptions rised on the RequestOperation op to propogate")
    def test_new_op_raises_base_exception(self, stage, op, mocker, arbitrary_base_exception):
        stage.next._execute_op = mocker.Mock(side_effect=arbitrary_base_exception)
        with pytest.raises(arbitrary_base_exception.__class__) as e_info:
            stage.run_op(op)
        assert op.callback.call_count == 0
        assert e_info.value is arbitrary_base_exception


@pytest.mark.describe(
    "CoordinateRequestAndResponseStage - .handle_pipeline_event() -- called with ResponseEvent"
)
class TestCoordinateRequestAndResponseSendIotRequestHandleEvent(StageTestBase):
    @pytest.fixture
    def op(self, mocker):
        return make_fake_request_and_response(mocker)

    @pytest.fixture
    def stage(self):
        return pipeline_stages_base.CoordinateRequestAndResponseStage()

    @pytest.fixture
    def iot_request(self, stage, op):
        stage.run_op(op)
        return stage.next.run_op.call_args[0][0]

    @pytest.fixture
    def iot_response(self, stage, iot_request):
        return pipeline_events_base.ResponseEvent(
            request_id=iot_request.request_id,
            status_code=fake_status_code,
            response_body=fake_response_body,
        )

    @pytest.mark.it(
        "Completes the RequestAndResponseOperation op with the matching request_id including response_body and status_code"
    )
    def test_completes_op_with_matching_request_id(self, stage, op, iot_response):
        stage.next.send_event_up(iot_response)
        assert_callback_succeeded(op=op)
        assert op.status_code == iot_response.status_code
        assert op.response_body == iot_response.response_body

    @pytest.mark.it(
        "Calls the unhandled error handler if there is no previous stage when request_id matches"
    )
    def test_matching_request_id_with_no_previous_stage(
        self, stage, op, iot_response, unhandled_error_handler
    ):
        stage.next.previous = None
        stage.next.send_event_up(iot_response)
        assert unhandled_error_handler.call_count == 1

    @pytest.mark.it(
        "Does nothing if an IotResponse with an identical request_id is received a second time"
    )
    def test_ignores_duplicate_request_id(self, stage, op, iot_response, unhandled_error_handler):
        stage.next.send_event_up(iot_response)
        assert_callback_succeeded(op=op)
        op.callback.reset_mock()

        stage.next.send_event_up(iot_response)
        assert op.callback.call_count == 0
        assert unhandled_error_handler.call_count == 0

    @pytest.mark.it(
        "Does nothing if an IotResponse with a request_id is received for an operation that returned failure"
    )
    def test_ignores_request_id_from_failure(
        self, stage, op, mocker, unhandled_error_handler, arbitrary_exception
    ):
        stage.next._execute_op = mocker.MagicMock(side_effect=arbitrary_exception)
        stage.run_op(op)

        req = stage.next.run_op.call_args[0][0]
        resp = pipeline_events_base.ResponseEvent(
            request_id=req.request_id,
            status_code=fake_status_code,
            response_body=fake_response_body,
        )

        op.callback.reset_mock()
        stage.next.send_event_up(resp)
        assert op.callback.call_count == 0
        assert unhandled_error_handler.call_count == 0

    @pytest.mark.it("Does nothing if an IotResponse with an unknown request_id is received")
    def test_ignores_unknown_request_id(self, stage, op, iot_response, unhandled_error_handler):
        iot_response.request_id = fake_request_id
        stage.next.send_event_up(iot_response)
        assert op.callback.call_count == 0
        assert unhandled_error_handler.call_count == 0


"""
A note on terms in the OpTimeoutStage tests:
    No-timeout ops are ops that don't need a timeout check
    Yes-timeout ops are ops that do need a timeout check
"""
timeout_intervals = {
    pipeline_ops_mqtt.MQTTSubscribeOperation: 10,
    pipeline_ops_mqtt.MQTTUnsubscribeOperation: 10,
}
yes_timeout_ops = list(timeout_intervals.keys())
no_timeout_ops = all_except(all_common_ops, yes_timeout_ops)

pipeline_stage_test.add_base_pipeline_stage_tests(
    cls=pipeline_stages_base.OpTimeoutStage,
    module=this_module,
    all_ops=all_common_ops,
    handled_ops=yes_timeout_ops,
    all_events=all_common_events,
    handled_events=[],
    extra_initializer_defaults={"timeout_intervals": timeout_intervals},
)


@pytest.fixture()
def mock_timer(mocker):
    return mocker.patch(
        "azure.iot.device.common.pipeline.pipeline_stages_base.Timer", autospec=True
    )


@pytest.mark.describe("OpTimeoutStage - run_op()")
class TestOpTimeoutStageRunOp(StageTestBase):
    @pytest.fixture(params=yes_timeout_ops)
    def yes_timeout_op(self, request, mocker):
        op = make_mock_op_or_event(request.param)
        op.callback = mocker.MagicMock()
        return op

    @pytest.fixture(params=no_timeout_ops)
    def no_timeout_op(self, request, mocker):
        op = make_mock_op_or_event(request.param)
        op.callback = mocker.MagicMock()
        return op

    @pytest.fixture
    def stage(self):
        return pipeline_stages_base.OpTimeoutStage()

    @pytest.mark.it("Sends ops that don't need a timer to the next stage")
    def test_sends_no_timer_op_down(self, stage, mock_timer, no_timeout_op):
        stage.run_op(no_timeout_op)
        assert stage.next.run_op.call_count == 1
        assert stage.next.run_op.call_args[0][0] == no_timeout_op

    @pytest.mark.it("Sends ops that do need a timer to the next stage")
    def test_sends_yes_timer_op_down(self, stage, mock_timer, yes_timeout_op):
        stage.run_op(yes_timeout_op)
        assert stage.next.run_op.call_count == 1
        assert stage.next.run_op.call_args[0][0] == yes_timeout_op

    @pytest.mark.it("Does not set a timer for ops that don't need a timer set")
    def test_does_not_set_timer(self, stage, mock_timer, no_timeout_op):
        stage.run_op(no_timeout_op)
        assert mock_timer.call_count == 0

    @pytest.mark.it("Set a timer for ops that need a timer set")
    def test_sets_timer(self, stage, mock_timer, yes_timeout_op):
        stage.run_op(yes_timeout_op)
        assert mock_timer.call_count == 1

    @pytest.mark.it("Starts the timer based on the timeout interval")
    def test_uses_timeout_interval(self, stage, mock_timer, yes_timeout_op):
        stage.run_op(yes_timeout_op)
        assert mock_timer.call_args[0][0] == timeout_intervals[yes_timeout_op.__class__]
        assert mock_timer.return_value.start.call_count == 1
        assert yes_timeout_op.timeout_timer == mock_timer.return_value

    @pytest.mark.it("Clears the timer when the op completes successfully")
    def test_clears_timer_on_success(self, stage, mock_timer, yes_timeout_op, next_stage_succeeds):
        stage.run_op(yes_timeout_op)
        assert mock_timer.return_value.cancel.call_count == 1
        assert getattr(yes_timeout_op, "timeout_timer", None) is None

    @pytest.mark.it("Clears the timer when the op fails with an arbitrary exception")
    def test_clears_timer_on_arbitrary_exception(
        self, stage, mock_timer, yes_timeout_op, next_stage_raises_arbitrary_exception
    ):
        stage.run_op(yes_timeout_op)
        assert mock_timer.return_value.cancel.call_count == 1
        assert getattr(yes_timeout_op, "timeout_timer", None) is None

    @pytest.mark.it("Does not clear the timer when the op fails with an arbitrary base exception")
    def test_doesnt_clear_timer_on_arbitrary_base_exception(
        self,
        stage,
        mock_timer,
        yes_timeout_op,
        next_stage_raises_arbitrary_base_exception,
        arbitrary_base_exception,
    ):
        with pytest.raises(arbitrary_base_exception.__class__):
            stage.run_op(yes_timeout_op)
        assert mock_timer.return_value.cancel.call_count == 0
        assert yes_timeout_op.timeout_timer == mock_timer.return_value

    @pytest.mark.it("Clears the timer when the op times out")
    def test_clears_timer_on_timeout(self, stage, mock_timer, yes_timeout_op):
        stage.run_op(yes_timeout_op)
        assert yes_timeout_op.timeout_timer == mock_timer.return_value
        timer_callback = mock_timer.call_args[0][1]
        timer_callback()
        assert getattr(yes_timeout_op, "timeout_timer", None) is None

    @pytest.mark.it("Calls the original callback with no error when the op completes with no error")
    def test_calls_callback_on_success(
        self, stage, mock_timer, yes_timeout_op, next_stage_succeeds
    ):
        stage.run_op(yes_timeout_op)
        assert_callback_succeeded(op=yes_timeout_op)

    @pytest.mark.it(
        "Calls the original callback with error when the op fails with an arbitrary exception"
    )
    def test_calls_callback_on_arbitrary_exception(
        self,
        stage,
        mock_timer,
        yes_timeout_op,
        next_stage_raises_arbitrary_exception,
        arbitrary_exception,
    ):
        stage.run_op(yes_timeout_op)
        assert_callback_failed(op=yes_timeout_op, error=arbitrary_exception)

    @pytest.mark.it(
        "Does not call the original callback when the op fails with an an arbitrary base exception"
    )
    def test_calls_callback_on_arbitrary_base_exception(
        self,
        stage,
        mock_timer,
        yes_timeout_op,
        next_stage_raises_arbitrary_base_exception,
        arbitrary_base_exception,
    ):
        callback = yes_timeout_op.callback  # capture before run_op because it changes inside run_op
        with pytest.raises(arbitrary_base_exception.__class__):
            stage.run_op(yes_timeout_op)
        assert callback.call_count == 0

    @pytest.mark.it("Calls the original callback with a PipelineTimeoutError when the op times out")
    def test_calls_callback_on_timeout(self, stage, mock_timer, yes_timeout_op):
        stage.run_op(yes_timeout_op)
        timer_callback = mock_timer.call_args[0][1]
        timer_callback()
        assert_callback_failed(op=yes_timeout_op, error=pipeline_exceptions.PipelineTimeoutError)


"""
A note on terms in the RetryStage tests:
    No-retry ops are ops that will never be retried.
    Yes-retry ops are ops that might be retired, depending on the error.
    Retry errors are errors that cause a retry for yes-retry ops
    Arbitrary errors will never cause a retry
"""

retry_intervals = {
    pipeline_ops_mqtt.MQTTSubscribeOperation: 20,
    pipeline_ops_mqtt.MQTTUnsubscribeOperation: 20,
    pipeline_ops_base.ConnectOperation: 20,
    pipeline_ops_mqtt.MQTTPublishOperation: 20,
}
yes_retry_ops = list(retry_intervals.keys())
no_retry_ops = all_except(all_common_ops, yes_retry_ops)
retry_errors = [pipeline_exceptions.PipelineTimeoutError]

pipeline_stage_test.add_base_pipeline_stage_tests(
    cls=pipeline_stages_base.RetryStage,
    module=this_module,
    all_ops=all_common_ops,
    handled_ops=[],
    all_events=all_common_events,
    handled_events=[],
    extra_initializer_defaults={"retry_intervals": retry_intervals, "ops_waiting_to_retry": []},
)


class RetryStageTestOpSend(object):
    """
    Tests for RetryStage to verify that ops get sent down
    """

    @pytest.fixture(params=no_retry_ops)
    def no_retry_op(self, request, mocker):
        op = make_mock_op_or_event(request.param)
        op.callback = mocker.MagicMock()
        return op

    @pytest.fixture(params=yes_retry_ops)
    def yes_retry_op(self, request, mocker):
        op = make_mock_op_or_event(request.param)
        op.callback = mocker.MagicMock()
        return op

    @pytest.mark.it("Sends ops that don't need retry to the next stage")
    def test_sends_no_retry_op_down(self, stage, no_retry_op):
        stage.run_op(no_retry_op)
        assert stage.next.run_op.call_count == 1
        assert stage.next.run_op.call_args[0][0] == no_retry_op

    @pytest.mark.it("Sends ops that do need retry to the next stage")
    def test_sends_yes_retry_op_down(self, stage, yes_retry_op):
        stage.run_op(yes_retry_op)
        assert stage.next.run_op.call_count == 1
        assert stage.next.run_op.call_args[0][0] == yes_retry_op


class RetryStageTestNoRetryOpCallback(object):
    """
    Tests for RetryStage for callbacks with no-retry ops.
    """

    @pytest.fixture(params=retry_errors)
    def retry_error(self, request):
        return request.param()

    @pytest.mark.it(
        "Calls the op callback with no error when an op that doesn't need retry succeeds"
    )
    def test_calls_callback_on_no_retry_op_success(self, stage, no_retry_op, next_stage_succeeds):
        stage.run_op(no_retry_op)
        assert_callback_succeeded(op=no_retry_op)

    @pytest.mark.it(
        "Calls the op callback with the correct error when an op that doesn't need retry fail with an arbitrary error"
    )
    def test_calls_callback_on_no_retry_op_arbitrary_exception(
        self, stage, no_retry_op, next_stage_raises_arbitrary_exception, arbitrary_exception
    ):
        stage.run_op(no_retry_op)
        assert_callback_failed(op=no_retry_op, error=arbitrary_exception)

    @pytest.mark.it(
        "Calls the op callback with the correct error when an op that doesn't need retry fail with a retry error"
    )
    def test_calls_callback_on_no_retry_op_retry_error(self, stage, no_retry_op, retry_error):
        stage.run_op(no_retry_op)
        stage.next.complete_op(op=no_retry_op, error=retry_error)
        assert_callback_failed(op=no_retry_op, error=retry_error)


class RetryStageTestNoRetryOpSetTimer(object):
    """
    Tests for RetryStage for not setting a timer for no-retry ops
    """

    @pytest.mark.it("Does not set a retry timer when an op that doesn't need retry succeeds")
    def test_no_timer_on_no_retry_op_success(
        self, stage, no_retry_op, next_stage_succeeds, mock_timer
    ):
        stage.run_op(no_retry_op)
        assert mock_timer.call_count == 0

    @pytest.mark.it(
        "Does not set a retry timer when an op that doesn't need retry fail with an arbitrary error"
    )
    def test_no_timer_on_no_retry_op_arbitrary_exception(
        self, stage, no_retry_op, next_stage_raises_arbitrary_exception, mock_timer
    ):
        stage.run_op(no_retry_op)
        assert mock_timer.call_count == 0

    @pytest.mark.it(
        "Does not set a retry timer when an op that doesn't need retry fail with a retry error"
    )
    def test_no_timer_on_no_retry_op_retry_error(self, stage, no_retry_op, retry_error, mock_timer):
        stage.run_op(no_retry_op)
        stage.next.complete_op(op=no_retry_op, error=retry_error)
        assert mock_timer.call_count == 0


class RetryStageTestYesRetryOpCallback(object):
    """
    Tests for RetryStage for callbacks with yes-retry ops
    """

    @pytest.mark.it("Calls the op callback with no error when an op that need retry succeeds")
    def test_callback_on_yes_retry_op_success(self, stage, yes_retry_op, next_stage_succeeds):
        stage.run_op(yes_retry_op)
        assert_callback_succeeded(op=yes_retry_op)

    @pytest.mark.it(
        "Calls the op callback with error when an op that need retry fails with an arbitrary error"
    )
    def test_callback_on_yes_retry_op_arbitrary_exception(
        self, stage, yes_retry_op, next_stage_raises_arbitrary_exception, arbitrary_exception
    ):
        stage.run_op(yes_retry_op)
        assert_callback_failed(op=yes_retry_op, error=arbitrary_exception)

    @pytest.mark.it(
        "Does not call the op callback when an op that need retry fail with a retry error"
    )
    def test_no_callback_on_yes_retry_op_retry_error(
        self, stage, yes_retry_op, retry_error, mock_timer
    ):
        stage.run_op(yes_retry_op)
        stage.next.complete_op(op=yes_retry_op, error=retry_error)
        assert yes_retry_op.callback.call_count == 0


class RetryStageTestYesRetryOpSetTimer(object):
    """
    Tests for RetryStage for setting or not setting timers for yes-retry ops
    """

    @pytest.mark.it("Does not set a retry timer when an op that need retry succeeds")
    def test_no_timer_on_yes_retry_op_success(
        self, stage, yes_retry_op, next_stage_succeeds, mock_timer
    ):
        stage.run_op(yes_retry_op)
        assert mock_timer.call_count == 0

    @pytest.mark.it(
        "Does not set a retry timer when an op that need retry fail with an arbitrary error"
    )
    def test_no_timer_on_yes_retry_op_arbitrary_exception(
        self, stage, yes_retry_op, next_stage_raises_arbitrary_exception, mock_timer
    ):
        stage.run_op(yes_retry_op)
        assert mock_timer.call_count == 0

    @pytest.mark.it("Sets a retry timer when an op that need retry fail with retry error")
    def test_yes_timer_on_yes_retry_op_retry_error(
        self, stage, yes_retry_op, retry_error, mock_timer
    ):
        stage.run_op(yes_retry_op)
        stage.next.complete_op(op=yes_retry_op, error=retry_error)
        assert mock_timer.call_count == 1

    @pytest.mark.it("Uses the correct timout when setting a retry timer")
    def test_uses_correct_timer_interval(self, stage, yes_retry_op, retry_error, mock_timer):
        stage.run_op(yes_retry_op)
        stage.next.complete_op(op=yes_retry_op, error=retry_error)
        assert mock_timer.call_args[0][0] == retry_intervals[yes_retry_op.__class__]


class RetryStageTestResubmitOp(object):
    """
    Tests for RetryStage for resubmiting ops for retry
    """

    @pytest.mark.it("Retries an op that needs retry after the retry interval elapses")
    def test_resubmits_after_retry_interval_elapses(
        self, stage, yes_retry_op, retry_error, mock_timer
    ):
        stage.run_op(yes_retry_op)
        assert stage.next.run_op.call_count == 1
        stage.next.run_op.reset_mock()
        stage.next.complete_op(op=yes_retry_op, error=retry_error)
        timer_callback = mock_timer.call_args[0][1]
        timer_callback()
        assert stage.next.run_op.call_count == 1
        assert stage.next.run_op.call_args[0][0] == yes_retry_op

    @pytest.mark.it("Clears the complete attribute on the op when retrying")
    def test_clears_complete_attribute_before_resubmitting(
        self, stage, yes_retry_op, retry_error, mock_timer
    ):
        stage.run_op(yes_retry_op)
        stage.next.complete_op(op=yes_retry_op, error=retry_error)
        assert yes_retry_op.completed
        timer_callback = mock_timer.call_args[0][1]
        timer_callback()
        assert not yes_retry_op.completed

    @pytest.mark.it("Clears the retry timer attribute on the op when retrying")
    def test_clears_retry_timer_before_retrying(self, stage, yes_retry_op, retry_error, mock_timer):
        stage.run_op(yes_retry_op)
        stage.next.complete_op(op=yes_retry_op, error=retry_error)
        assert yes_retry_op.retry_timer
        timer_callback = mock_timer.call_args[0][1]
        timer_callback()
        assert getattr(yes_retry_op, "retry_timer", None) is None


class RetryStageTestResubmitedOpCompletion(object):
    """
    Tests for RetryStage for resubmitted op completion
    """

    @pytest.mark.it("Calls the original callback with success when the retried op succeeds")
    def test_calls_callback_on_retried_op_success(
        self, stage, yes_retry_op, retry_error, mock_timer
    ):
        op_callback = yes_retry_op.callback
        stage.run_op(yes_retry_op)
        stage.next.complete_op(op=yes_retry_op, error=retry_error)
        timer_callback = mock_timer.call_args[0][1]
        timer_callback()
        assert op_callback.call_count == 0
        stage.next.complete_op(op=yes_retry_op)
        assert yes_retry_op.callback == op_callback
        assert_callback_succeeded(op=yes_retry_op)

    @pytest.mark.it(
        "Calls the original callback with error when the retried op compltes with an arbitrary error"
    )
    def test_calls_callback_on_retried_op_arbitrary_exception(
        self, stage, yes_retry_op, retry_error, mock_timer, arbitrary_exception, mocker
    ):

        stage.run_op(yes_retry_op)
        stage.next.complete_op(op=yes_retry_op, error=retry_error)
        timer_callback = mock_timer.call_args[0][1]
        timer_callback()
        stage.next.complete_op(op=yes_retry_op, error=arbitrary_exception)
        assert_callback_failed(op=yes_retry_op, error=arbitrary_exception)

    @pytest.mark.it(
        "Does not calls the original callback with error when the retried op compltes with an retry error"
    )
    def test_no_callback_on_retried_op_retry_error(
        self, stage, yes_retry_op, retry_error, mock_timer
    ):
        op_callback = yes_retry_op.callback
        stage.run_op(yes_retry_op)
        stage.next.complete_op(op=yes_retry_op, error=retry_error)
        timer_callback = mock_timer.call_args[0][1]
        timer_callback()
        stage.next.complete_op(op=yes_retry_op, error=retry_error)
        assert op_callback.call_count == 0

    @pytest.mark.it("Sets a new retry timer error when the retried op compltes with an retry error")
    def test_sets_timer_on_retried_op_retry_error(
        self, stage, yes_retry_op, retry_error, mock_timer
    ):
        stage.run_op(yes_retry_op)
        stage.next.complete_op(op=yes_retry_op, error=retry_error)
        assert mock_timer.call_count == 1
        timer_callback = mock_timer.call_args[0][1]
        timer_callback()
        stage.next.complete_op(op=yes_retry_op, error=retry_error)
        assert mock_timer.call_count == 2


@pytest.mark.describe("RetryStage - run_op()")
class TestRetryStageRunOp(
    StageTestBase,
    RetryStageTestOpSend,
    RetryStageTestNoRetryOpCallback,
    RetryStageTestNoRetryOpSetTimer,
    RetryStageTestYesRetryOpCallback,
    RetryStageTestYesRetryOpSetTimer,
    RetryStageTestResubmitOp,
    RetryStageTestResubmitedOpCompletion,
):
    @pytest.fixture
    def stage(self):
        return pipeline_stages_base.RetryStage()


pipeline_stage_test.add_base_pipeline_stage_tests(
    cls=pipeline_stages_base.ReconnectStage,
    module=this_module,
    all_ops=all_common_ops,
    handled_ops=[],
    all_events=all_common_events,
    handled_events=[],
    extra_initializer_defaults={
        "reconnect_timer": None,
        "virtually_connected": False,
        "reconnect_delay": 10,
    },
)


@pytest.mark.describe("ReconnectStage - .run_op()")
class TestReconnectStageRunOp(StageTestBase):
    @pytest.fixture
    def stage(self):
        return pipeline_stages_base.ReconnectStage()

    @pytest.mark.it(
        "Sets the stage virtually_connected attribute to True when a ConnectOperation is sent down"
    )
    def test_connect_op_virtual_connection(self, mocker, stage):
        op = pipeline_ops_base.ConnectOperation(mocker.MagicMock())
        assert not stage.virtually_connected
        stage.run_op(op)
        assert stage.virtually_connected

    @pytest.mark.it(
        "Keeps the stage virtually_connected attribute set to True, even if the ConnectOperation fails"
    )
    def test_connect_op_virtual_connection_failure(self, mocker, stage, arbitrary_exception):
        op = pipeline_ops_base.ConnectOperation(mocker.MagicMock())
        stage.next._execute_op = mocker.MagicMock(side_effect=arbitrary_exception)
        stage.run_op(op)
        assert op.callback.call_count == 1
        assert op.callback.call_args[1]["error"] == arbitrary_exception
        assert stage.virtually_connected

    @pytest.mark.it(
        "Sets the stage virtually_connected attribute to False when a DisconnectOperation is sent down"
    )
    def test_disconnect_op_virtual_connection(self, mocker, stage):
        op = pipeline_ops_base.DisconnectOperation(mocker.MagicMock())
        stage.virtually_connected = True
        stage.run_op(op)
        assert not stage.virtually_connected

    @pytest.mark.it(
        "Keeps the stage virtually_connected attribute set to False, even if the DisconnectOperation fails"
    )
    def test_disconnect_op_virtual_connection_failure(self, mocker, stage, arbitrary_exception):
        op = pipeline_ops_base.DisconnectOperation(mocker.MagicMock())
        stage.virtually_connected = True
        stage.next._execute_op = mocker.MagicMock(side_effect=arbitrary_exception)
        stage.run_op(op)
        assert op.callback.call_count == 1
        assert op.callback.call_args[1]["error"] == arbitrary_exception
        assert not stage.virtually_connected


@pytest.mark.describe("ReconnectStage - .on_connected()")
class TestReconnectStageOnConnected(StageTestBase):
    @pytest.fixture
    def stage(self):
        return pipeline_stages_base.ReconnectStage()

    @pytest.mark.it("Clears any reconnect timers that may be running")
    def test_clears_reconnect_timer(self, stage, mock_timer):
        timer = mock_timer.return_value
        stage.reconnect_timer = timer
        assert timer.cancel.call_count == 0
        stage.on_connected()
        assert stage.reconnect_timer is None
        assert mock_timer.return_value.cancel.call_count == 1

    @pytest.mark.it("Does not fail if there there are no pending reconnect timers")
    def test_succeeds_if_no_reconnect_timer(self, stage, mock_timer):
        stage.on_connected()
        assert mock_timer.return_value.cancel.call_count == 0

    @pytest.mark.it("Calls the previous stage on_connected handler if there was a pending timer")
    def test_calls_previous_stage_on_connected(self, mocker, stage, mock_timer):
        mocker.spy(stage.previous, "on_connected")
        stage.reconnect_timer = mock_timer.return_value
        assert stage.previous.on_connected.call_count == 0
        stage.on_connected()
        assert stage.previous.on_connected.call_count == 1

    @pytest.mark.it(
        "Calls the previous stage on_connected handler if there was not a pending_timer"
    )
    def test_calls_previous_stage_on_connected_without_timer(self, mocker, stage):
        mocker.spy(stage.previous, "on_connected")
        assert stage.previous.on_connected.call_count == 0
        stage.on_connected()
        assert stage.previous.on_connected.call_count == 1


default_reconnect_time = 10


@pytest.mark.describe("ReconnectStage - .on_disconnected()")
class TestReconnectStageOnDisconnected(StageTestBase):
    @pytest.fixture
    def stage(self):
        return pipeline_stages_base.ReconnectStage()

    @pytest.mark.it(
        "Clears a previous reconnect timer if the stage virtually_connected attribute is set to True"
    )
    def test_clears_previous_reconnect_timer_if_virtually_connected(self, stage, mock_timer):
        timer = mock_timer.return_value
        stage.reconnect_timer = timer
        stage.virtually_connected = True
        stage.on_disconnected()
        stage.reconnect_timer is None
        timer.cancel.call_count == 1

    @pytest.mark.it(
        "Sets a reconnect timer if the stage virtually_connected attribute is set to True"
    )
    def test_sets_new_reconnect_timer_if_virtually_connected(self, stage, mock_timer):
        stage.virtually_connected = True
        stage.pipeline_root.connected = True
        stage.on_disconnected()
        assert stage.reconnect_timer == mock_timer.return_value
        assert mock_timer.call_count == 1
        assert mock_timer.call_args[0][0] == default_reconnect_time

    @pytest.mark.it(
        "Does not set a reconnect timer if the stage virtually_connected attribute is set to False"
    )
    def test_does_not_set_reconnect_timer_if_not_virtually_connected(self, stage, mock_timer):
        stage.on_disconnected()
        assert stage.reconnect_timer is None
        assert mock_timer.call_count == 0

    @pytest.mark.it(
        "Calls the previous stage on_disconnected handler if the stage virtually_connected attribute is set to True"
    )
    def test_calls_previous_stage_on_disconnected_virtually_connected(
        self, mocker, stage, mock_timer
    ):
        mocker.spy(stage.previous, "on_disconnected")
        stage.pipeline_root.connected = True
        stage.virtually_connected = True
        stage.on_disconnected()
        assert stage.previous.on_disconnected.call_count == 1

    @pytest.mark.it(
        "Calls the previous stage on_disconnected handler if the stage virtually_connected attribute is set to False"
    )
    def test_calls_previous_stage_on_disconnected_not_virtually_connected(
        self, mocker, stage, mock_timer
    ):
        mocker.spy(stage.previous, "on_disconnected")
        stage.on_disconnected()
        assert stage.previous.on_disconnected.call_count == 1


class ArbitraryException(Exception):
    pass


@pytest.mark.describe("ReconnectStage - reconnect timer routine")
class TestReconnectStageReconnectTimerRoutine(StageTestBase):
    @pytest.fixture
    def stage(self):
        return pipeline_stages_base.ReconnectStage()

    @pytest.fixture
    def timer_routine(self, stage, mock_timer):
        stage.virtually_connected = True
        stage.pipeline_root.connected = True
        stage.on_disconnected()
        return mock_timer.call_args[0][1]

    @pytest.mark.it("Runs in the pipeline thread with nowait")
    def test_invokes_on_pipeline_thread(
        self, mocker, stage, timer_routine, fake_non_pipeline_thread
    ):
        done = threading.Event()
        stage.pipeline_root.connected = False
        names = {"thread_name": None}

        def save_thread_name(*args, **kwargs):
            names["thread_name"] = threading.current_thread().name
            done.set()

        logger = mocker.patch.object(pipeline_stages_base, "logger")
        logger.debug.side_effect = save_thread_name

        assert threading.current_thread().name != "pipeline"
        timer_routine()
        done.wait()
        assert names["thread_name"] == "pipeline"

    @pytest.mark.it("Does not send a ConnectOperation down if the transport is already connected")
    def test_when_connected_does_not_send_connect_operation_down(self, stage, timer_routine):
        stage.pipeline_root.connected = True
        timer_routine()
        assert stage.next.run_op.call_count == 0

    @pytest.mark.it("Sends a ConnectOperation down if the transport is not connected")
    def test_when_disconnected_sends_connect_operation_down(self, stage, timer_routine):
        stage.pipeline_root.connected = False
        timer_routine()
        assert stage.next.run_op.call_count == 1
        assert type(stage.next.run_op.call_args[0][0]) == pipeline_ops_base.ConnectOperation

    @pytest.mark.parametrize(
        "error_class",
        [
            ArbitraryException,
            transport_exceptions.UnauthorizedError,
            transport_exceptions.ProtocolClientError,
            pipeline_exceptions.PipelineError,
        ],
    )
    @pytest.mark.it(
        "Does not sets a new reconnect timer if the ConnectOperation fails with an error that probably isn't transient"
    )
    def test_does_not_set_new_timer_on_non_transient_connect_error(
        self, stage, mock_timer, timer_routine, error_class
    ):
        stage.pipeline_root.connected = False
        assert mock_timer.call_count == 1
        timer_routine()
        assert stage.next.run_op.call_count == 1
        op = stage.next.run_op.call_args[0][0]
        assert type(op) == pipeline_ops_base.ConnectOperation
        op.callback(op, error=error_class())
        assert mock_timer.call_count == 1

    @pytest.mark.parametrize(
        "error_class",
        [
            pipeline_exceptions.OperationCancelled,
            pipeline_exceptions.PipelineTimeoutError,
            transport_exceptions.ConnectionFailedError,
            transport_exceptions.ConnectionDroppedError,
        ],
    )
    @pytest.mark.it(
        "Sets a new reconnect timer if the ConnectOperation fails with an error that might be transient"
    )
    def test_sets_new_timer_on_transient_connect_error(
        self, stage, mock_timer, timer_routine, error_class
    ):
        stage.pipeline_root.connected = False
        assert mock_timer.call_count == 1
        timer_routine()
        assert stage.next.run_op.call_count == 1
        op = stage.next.run_op.call_args[0][0]
        assert type(op) == pipeline_ops_base.ConnectOperation
        op.callback(op, error=error_class())
        assert mock_timer.call_count == 2
